from hops import elasticsearch
from elasticsearch import Elasticsearch
import warnings
from elasticsearch.helpers import scan
import pandas as pd
import json

class ElasticIndex(object):

    def __init__(self, index, sort='_id'):
        warnings.filterwarnings("ignore")
        self.index = index
        self.sort = sort
        self.es_config = self._get_es_config()
        self.qualified_index = self.es_config['es.resource']
        self.auth_token = elasticsearch.get_authorization_token()
        self.header = {'Content-Type' : 'application/json; charset=utf8','authorization' : self.auth_token}
        self.host = self._get_host()
        self.client = Elasticsearch(self.host, verify_certs=False)

    def _get_es_config(self):
        es_config = elasticsearch.get_elasticsearch_config(self.index)
        es_config['es.nodes'] = es_config['es.nodes'].replace('9300', '9200')
        return es_config

    def _get_host(self):
        if self._is_pyspark():
            return self.es_config['es.nodes']
        else:
            return "https://" + self.es_config['es.nodes']

    def _is_pyspark(self):
        try:
            import pyspark
            return True
        except:
            return False


    def iterator(self, max = None, query = {"query": {"match_all": {}}}, _source=None, size=1000):
        if _source == None:
            scan_result = scan(self.client, query, index=self.qualified_index, 
                               size=size, _source=None, headers=self.header, scroll_kwargs={'headers': self.header})
        else:
            scan_result = scan(self.client, query, index=self.qualified_index, 
                               size=size, _source=_source, headers=self.header, scroll_kwargs={'headers': self.header})

        i = 0
        for row in scan_result:
            if i == max: break
            i = i + 1
            yield row['_source']
            
        try: 
            scan_result.close()
            del scan_result
        except: 
            pass

    
    def get_by_ids(self, ids, ignore_not_found = False):
        response = self.client.mget(index = self.qualified_index, doc_type = '_doc', headers=self.header, body = {'ids': ids})
        if ignore_not_found:
            return [doc['_source'] for doc in response['docs'] if  doc['found']]
        else:
            return [doc['_source'] if doc['found'] else None for doc in response['docs']]
        
    def __len__(self):
        response = self.client.count(index = self.qualified_index,  headers=self.header)
        return response['count']
    
    def _get_start_and_rows(self, idx):
        if type(idx) is int:
            start = idx if idx >= 0 else 0
            rows = 1
        elif type(idx) is slice:
            start = idx.start if idx.start >= 0 else 0
            rows = idx.stop-(idx.start - 1)
        return start, rows
    
    def get_rows(self, idx):
        start, rows = self._get_start_and_rows(idx)

        response = self.client.search(index=self.qualified_index, headers=self.header, body={
            "from": start,
            "size": rows,
            "sort": [{self.sort: { "order": "asc" } }],
            "query": {"match_all": {}}})

        return [hit['_source'] for hit in response['hits']['hits']], rows
    
    def __getitem__(self, idx):
        result, rows = self.get_rows(idx)
        if rows == 1: return result[0]
        return result
    

    def post_to_elastic(self, json_batch_string, index_type='_doc'):
        response = requests.post(self.host + "/" + self.qualified_index +'/'+ index_type + '/_bulk',
                                 data=json_batch_string, headers=self.headers, timeout=5)
        return response

    def df_to_json(self, records, id):
        json_batch_string = ''
        for index, record in records.iterrows():
            metadata = json.dumps({'index': {'_id': record[id]}})
            record = record.drop(labels=[id])
            doc = record.to_json()
            json_batch_string += str(metadata) + '\n' + doc + '\n'
        return json_batch_string

    def list_of_dicts_to_json(self, records, id):
        json_batch_string = ''
        for record in records:
            metadata = json.dumps({'index': {'_id': record[id]}})
            doc = json.dumps(record)
            json_batch_string += str(metadata) + '\n' + doc + '\n'
        return json_batch_string

    def to_json(self, records, id):
        if isinstance(records, pd.DataFrame):
            return self.df_to_json(records, id)
        else:
            return self.list_of_dicts_to_json(records, id)

    def to_elastic(self, records, id='uuid', index_type='_doc'):
        response = self.post_to_elastic(self.to_json(records, id), index_type)
        json_response = json.loads(response.text)
        if json_response and json_response['errors'] != False: print(json_response)
        return response
