import tensorflow as tf
from hops import featurestore
from tensorflow.keras.layers.experimental import preprocessing
from hops import hdfs

def get_tf_record_schema(dataset_path):
    tf_record_schema = featurestore.get_training_dataset_tf_record_schema(dataset_path)
    for key, value  in tf_record_schema.items():
        if isinstance(value, tf.io.VarLenFeature):
            tf_record_schema[key] = tf.io.FixedLenFeature([], tf.string, default_value='')
    return tf_record_schema

def get_parse_tfrecord_batch_fn(tf_record_schema):
    def parse_tfrecord_batch(batch):
        return tf.io.parse_example(batch, tf_record_schema)
    return parse_tfrecord_batch

def get_dataset_dir(dataset_name, version=1, split=None):
    proj_name = hdfs.project_name()
    if split is not None:
        dataset_dir = "{}{}_Training_Datasets/{}_{}/{}".format(hdfs.project_path(), proj_name, dataset_name, version, split)
    else:
        dataset_dir = "{}{}_Training_Datasets/{}_{}".format(hdfs.project_path(), proj_name, dataset_name, version)
    return dataset_dir

def get_dataset(dataset_path, batch_size=32):
    dataset_dir = featurestore.get_training_dataset_path(dataset_path)
    input_files = tf.io.gfile.glob(dataset_dir + "/part-r-*")
    dataset = tf.data.TFRecordDataset(input_files)
    
    tf_record_schema = get_tf_record_schema(dataset_path)
    parse_tfrecord_batch = get_parse_tfrecord_batch_fn(tf_record_schema)
    
    return dataset.batch(batch_size).map(parse_tfrecord_batch)

def get_tf_split_dataset(dataset_name,split,tf_record_schema, batch_size=32,version=1):
    dataset_path = get_dataset_dir(dataset_name,version,split)
    input_files = tf.io.gfile.glob(dataset_path + "/part-r-*")
    dataset = tf.data.TFRecordDataset(input_files)
    parse_tfrecord_batch = get_parse_tfrecord_batch_fn(tf_record_schema)
    #tfutils.get_dataset(dataset_path,batch_size) # Take schema from featurestore (not available using new API)
    return dataset.batch(batch_size).map(parse_tfrecord_batch)

def get_category_encoding_layer(name, dataset, dtype, max_tokens=None):
    # Create a StringLookup layer which will turn strings into integer indices
    if dtype == 'string':
        index = preprocessing.StringLookup(max_tokens=max_tokens)
    else:
        index = preprocessing.IntegerLookup(max_values=max_tokens)

    # Prepare a Dataset that only yields our feature
    feature_ds = dataset.map(lambda x, y: x[name])

    # Learn the set of possible values and assign them a fixed integer index.
    index.adapt(feature_ds)

    # Create a Discretization for our integer indices.
    encoder = preprocessing.CategoryEncoding(max_tokens=index.vocab_size())

    # Prepare a Dataset that only yields our feature.
    feature_ds = feature_ds.map(index)

    # Learn the space of possible indices.
    encoder.adapt(feature_ds)

    # Apply one-hot encoding to our indices. The lambda function captures the
    # layer so we can use them, or include them in the functional model later.
    return lambda feature: encoder(index(feature))

def get_categorical_features(feature_names, dataset):
    all_inputs = []
    encoded_features = []
    for feature_name in feature_names:
        categorical_col = tf.keras.Input(shape=(1,), name=feature_name, dtype='string')
        encoding_layer = get_category_encoding_layer(feature_name, dataset, dtype='string')
        encoded_categorical_col = encoding_layer(categorical_col)
        all_inputs.append(categorical_col)
        encoded_features.append(encoded_categorical_col)
        
    all_features = tf.keras.layers.concatenate(encoded_features)
    return all_features, all_inputs

def get_normalization_layer(name, dataset):
    # Create a Normalization layer for our feature.
    normalizer = preprocessing.Normalization()

    # Prepare a Dataset that only yields our feature.
    feature_ds = dataset.map(lambda x, y: x[name])

    # Learn the statistics of the data.
    normalizer.adapt(feature_ds)

    return normalizer

def get_normalized_features(feature_names, dataset):
    all_inputs = []
    encoded_features = []

    for feature_name in feature_names:
        numeric_col = tf.keras.Input(shape=(1,), name=feature_name)
        normalization_layer = get_normalization_layer(feature_name, dataset)
        encoded_numeric_col = normalization_layer(numeric_col)
        all_inputs.append(numeric_col)
        encoded_features.append(encoded_numeric_col)
        
    all_features = tf.keras.layers.concatenate(encoded_features)
    return all_features, all_inputs
